/*
 *  raytracer.cpp
 *
 *  Author : Ahmed Obied (ahmed.obied@gmail.com)
 *
 */

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Author : Ahmed Obied                                                      //
// Date   : April 20th 2004                                                  //
// Purpose: Create a distributed ray tracer                                  //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

// include all the libraries needed in the program
#include "iostream.h"
#include "fstream.h"
#include "raytracer.h"
#include "pixelmap.h"
#include "vector.h"
#include "sphere.h"
#include "shape.h"
#include "plane.h"
#include "math.h"
// define all the simulated fuzzy phenomena
// if the value is 1 then we simulate, else if the value is 0 then normal
// ray tracing will be done
// Effects 
#define MOTION_BLUR 0
#define BLURRED_REFLECTION 0
#define DEPTH_OF_FIELD 0
#define SOFT_SHADOWS 1
// displacement for motion blur
#define DISP 0.05
// time for motion blur
#define TIME 1
// width and height of final image
#define WIDTH 700
#define HEIGHT 700
// number of samples required 
#define SAMPLES 64
// depth for reflection 
#define MAX 1 
// jitter function to distribute rays within a pixel
#define RAND() (double)((rand()/(double)RAND_MAX)-0.5)
// date needed in the program
Color shadowColor;
PixelMap output(WIDTH,HEIGHT);
Color scene[WIDTH][HEIGHT];
Sphere* sphere_1 = new Sphere();
Sphere* sphere_2 = new Sphere();
Sphere* sphere_3 = new Sphere();
Sphere* lens = new Sphere();
Sphere* focal_plane = new Sphere();

//-----------------------------------------------------------------------------
// Parameters    : None
// Return value  : None
// Purpose       : The constructor is not used for any special purpose here. It
//                 only create a banner showing that the distributed ray 
//                 tracing has began
// Method called : RayTracer
//-----------------------------------------------------------------------------
RayTracer::RayTracer()
{
  // print the ray tracer banner
  cout<<"\n";
  cout<<"---------------------------\n";
  cout<<"Distributed ray tracing ...\n";
  cout<<"---------------------------\n";
}

//-----------------------------------------------------------------------------
// Parameters    : None
// Return value  : None
// Purpose       : This function is used to initialize everthing in the scene.
//                 It initializes the light, scene objects and start tracing.
// Method called : start
//-----------------------------------------------------------------------------
void RayTracer::start()
{
  // init data and scene
  initCamera();
  initLights();
  initScene();
  startTracing();
  outputImage();
  // init the shadow color
  shadowColor.red = 0;
  shadowColor.green = 0;
  shadowColor.blue = 0;
}

//-----------------------------------------------------------------------------
// Parameters    : None
// Return value  : None
// Purpose       : This function is used to initialize the camera data which
//                 includes the location and the lenses for the depth of field
//                 effect
// Method called : initCamera
//-----------------------------------------------------------------------------
void RayTracer::initCamera()
{    
  // set the eye location
  eye.assign(0,0,10);
  // set the background color
  background.red = 0;
  background.green = 0;
  background.blue = 0;
  // check if we want to simulate depth of field
  if (DEPTH_OF_FIELD)
    {
      gmVector3 temp(0.0,-0.8,-10);
      // set the lens radius and location
      lens->setRadius(0.5);
      lens->setCenter(eye);
      // set the focal plane location
      focal_plane->setRadius(1.0);
      focal_plane->setCenter(temp);
    }
}

//-----------------------------------------------------------------------------
// Parameters    : None
// Return value  : None
// Purpose       : This function is used to initialize the lights in the scene
//                 If the soft shadow effect is set then an area light is 
//                 being created to simulate soft shadows
// Method called : initLights
//-----------------------------------------------------------------------------
void RayTracer::initLights()
{
  gmVector3 temp,org;
  Light light;
  Color amb,dif,spec;
  // if we want a point light source
  if (!SOFT_SHADOWS)
    {
      org.assign(60,100,200);
      light.setPosition(org);
      amb.assign(1,1,1);
      dif.assign(1,1,1);
      spec.assign(1,1,1);
      light.setMat(amb,dif,spec);
      area_light.push_back(light);
    }
  // if we want an area light source
  if (SOFT_SHADOWS)
    {
      org.assign(60,100,200);
      light.setPosition(org);
      amb.assign(1,1,1);
      dif.assign(1,1,1);
      spec.assign(1,1,1);
      light.setMat(amb,dif,spec);
      area_light.push_back(light);
      // create a set of samples to simulate an area light source
      for ( int i = 0 ; i < SAMPLES ; i++)
	{
	  // the samples are randomly distributed around the center of the
	  // original light source
	  temp[0] = org[0] + (rand()%40);
	  temp[1] = org[1] + (rand()%40);
	  temp[2] = org[2] + (rand()%40);
	  light.setPosition(temp);
	  area_light.push_back(light);
	}
    }
}

//-----------------------------------------------------------------------------
// Parameters    : None
// Return value  : None
// Purpose       : This function is used to initialize the scene objects 
// Method called : initScene
//-----------------------------------------------------------------------------
void RayTracer::initScene()
{
  // data needed for settings
  gmVector3 temp,min,max;
  Color amb,dif,spec;
  // if we want to simulate motion blur, then load the the scene for it
  if (MOTION_BLUR)
    {
      // set the spheres and planes location and material 
      amb.assign(0.1,0.1,0.1);
      dif.assign(0.5,0.5,1);
      spec.assign(1,1,1);
      temp.assign(0.0,-0.2,-10);
      sphere_1->setRadius(0.8);
      sphere_1->setCenter(temp);
      sphere_1->setMat(amb,dif,spec,0,0,50);
      Plane* plane = new Plane();  
      temp.assign(0,1,0);
      min.assign(-1000,-1.1,-1000);
      max.assign(1000,-0.9,1000);
      amb.assign(0.1,0.1,0.1);
      dif.assign(0.8,0.8,0.8);
      spec.assign(1,1,1);
      plane->setDistance(1);
      plane->setNormal(temp);
      plane->setMin(min);
      plane->setMax(max);
      plane->setMat(amb,dif,spec,0,0,100);
      shapes.push_back(sphere_1);
      shapes.push_back(plane);
    }
  else
     // if we want to simulate blurred reflection, 
     // then load the the scene for it
    if (BLURRED_REFLECTION)
      {
	// set the spheres and planes locations
	amb.assign(0.1,0.1,0.1);
	dif.assign(0.5,1,0.5);
	spec.assign(1,1,1);
	temp.assign(0.0,-0.2,-10);
	sphere_1->setRadius(0.8);
	sphere_1->setCenter(temp);
	sphere_1->setMat(amb,dif,spec,0.8,0,50);
	Plane* plane = new Plane();  
	temp.assign(0,1,0);
	min.assign(-1000,-1.1,-1000);
	max.assign(1000,-0.9,1000);
	amb.assign(0.1,0.1,0.1);
	dif.assign(0.8,0.8,0.8);
	spec.assign(1,1,1);
	plane->setDistance(1);
	plane->setNormal(temp);
	plane->setMin(min);
	plane->setMax(max);
	plane->setMat(amb,dif,spec,0,0,100);
	shapes.push_back(sphere_1);
	shapes.push_back(plane);
      }
    else
      // if we want to simulate depth of field then load the scene for it
      if(DEPTH_OF_FIELD)
	{
	  // set the spheres locations and material
	  amb.assign(0.1,0.1,0.1);
	  dif.assign(0.5,0.5,1);
	  spec.assign(1,1,1);
	  temp.assign(0.0,-0.8,-10);
	  sphere_1->setRadius(0.8);
	  sphere_1->setCenter(temp);
	  sphere_1->setMat(amb,dif,spec,0,0,200);
	  shapes.push_back(sphere_1);
	  amb.assign(0.1,0.1,0.1);
	  dif.assign(0.5,0.5,1);
	  spec.assign(1,1,1);
	  temp.assign(0.0,0.8,-10);
	  sphere_2->setRadius(0.8);
	  sphere_2->setCenter(temp);
	  sphere_2->setMat(amb,dif,spec,0,0,200);
	  shapes.push_back(sphere_2);
	} 
      else
	if (SOFT_SHADOWS)
	  {
	    // set the spheres and planes location and material 
	    amb.assign(0.1,0.1,0.1);
	    dif.assign(0.5,0.5,1);
	    spec.assign(1,1,1);
	    temp.assign(0.0,-0.2,-10);
	    sphere_1->setRadius(0.8);
	    sphere_1->setCenter(temp);
	    sphere_1->setMat(amb,dif,spec,0,0,50);
	    Plane* plane = new Plane();  
	    temp.assign(0,1,0);
	    min.assign(-1000,-1.1,-1000);
	    max.assign(1000,-0.9,1000);
	    amb.assign(0.1,0.1,0.1);
	    dif.assign(0.8,0.8,0.8);
	    spec.assign(1,1,1);
	    plane->setDistance(1);
	    plane->setNormal(temp);
	    plane->setMin(min);
	    plane->setMax(max);
	    plane->setMat(amb,dif,spec,0,0,100);
	    shapes.push_back(sphere_1);
	    shapes.push_back(plane);
	  }
	else
	  {
	    /*
	      amb.assign(0.1,0.1,0.1);
	      dif.assign(1,1,1);
	      spec.assign(1,1,1);
	      temp.assign(-0.8,0.0,-10);
	      sphere_1->setRadius(0.8);
	      sphere_1->setCenter(temp);
	      sphere_1->setMat(amb,dif,spec,0.7,0,50);
	  
	  amb.assign(0.1,0.1,0.1);
	  dif.assign(1,1,1);
	  spec.assign(1,1,1);
	  temp.assign(0.8,0.0,-10);
	  sphere_2->setRadius(0.8);
	  sphere_2->setCenter(temp);
	  sphere_2->setMat(amb,dif,spec,0.7,0,50);
	  
	  Plane* plane = new Plane();  
	  temp.assign(0,1,0);
	  min.assign(-1000,-1.1,-1000);
	  max.assign(1000,-0.9,1000);
	  amb.assign(0.1,0.1,0.1);
	  dif.assign(0.8,0.8,0.8);
	  spec.assign(1,1,1);
	  plane->setDistance(1);
	  plane->setNormal(temp);
	  plane->setMin(min);
	  plane->setMax(max);
	  plane->setMat(amb,dif,spec,0,0,100);
	  
	  shapes.push_back(sphere_1);
	  shapes.push_back(sphere_2);
	  shapes.push_back(plane);
	  */
	  
	  amb.assign(0.1,0.1,0.1);
	  dif.assign(0.5,0.5,1);
	  spec.assign(1,1,1);
	  temp.assign(0.0,-0.2,-10);
	  sphere_1->setRadius(0.8);
	  sphere_1->setCenter(temp);
	  sphere_1->setMat(amb,dif,spec,0.7,0,50);
	  Plane* plane = new Plane();  
	  temp.assign(0,1,0);
	  min.assign(-0.5,-1.1,-20);
	  max.assign(0.5,-0.9,20);
	  amb.assign(1,1,1);
	  dif.assign(1,1,1);
	  spec.assign(1,1,1);
	  plane->setDistance(1);
	  plane->setNormal(temp);
	  plane->setMin(min);
	  plane->setMax(max);
	  plane->setMat(amb,dif,spec,0,0,100);
	  shapes.push_back(sphere_1);
	  shapes.push_back(plane);
	  
      

	}
}

//-----------------------------------------------------------------------------
// Parameters    : None
// Return value  : None
// Purpose       : This function is used to update the location of the object
//                 in order to simulate motion blur
// Method called : updateLocation
//-----------------------------------------------------------------------------
void RayTracer::updateLocation()
{
  gmVector3 temp;
  // if motion blur is set the update the object location
  if(MOTION_BLUR)
    {
      // I'm using a sphere to simulate its motion 
      temp = sphere_1->getCenter();
      // use a displacment to shift to the next frame
      temp[1] = temp[1] + DISP;
      sphere_1->setCenter(temp);
    }
}

//-----------------------------------------------------------------------------
// Parameters    : None
// Return value  : None
// Purpose       : This function is the heart of the distributed ray tracer.
//                 Rays are being distributed to simulate some of the fuzzy
//                 effects and to allow anti-aliasing
// Method called : startTracing
//-----------------------------------------------------------------------------
void RayTracer::startTracing()
{
  int status;
  gmVector3 focal_intersect;
  gmVector3 lens_point;
  gmVector3 a,b,c;
  // set the scene vectors 
  a.assign(1,1,-2.414);
  b.assign(-1,1,-2.414);
  c.assign(-1,-1,-2.414);
  Color pixelColor(0,0,0);
  pixelColor.red = 0;
  pixelColor.green = 0;
  pixelColor.blue = 0;
  // start tracing
  cout<<"Ray tracing starts ...\n";
  for ( int i = 0 ; i < WIDTH ; i++)
    for ( int j = 0 ; j < HEIGHT ; j++)
      // set the buffer scene to the background color
      scene[i][j] = background;
  // loop used to distribute rays in time to simulate motion blur
  for ( int k = 0 ; k < TIME ; k++)
    {
      // go over all the pixels in the scene
      for (int i = 0 ; i < WIDTH ; i++)
	for ( int j = 0 ; j < HEIGHT ; j++)
	  {
	    // compute the ray direction
	    rayDirection = (b-eye)
	      +(i*(a-b)/WIDTH)
	      +(j*(c-b)/HEIGHT);
	    rayDirection.normalize();
	    // if we wanto to distribute rays then we loop up to the number
	    // of sampels we have
	    for ( int k = 0 ; k < SAMPLES ; k++)
	      {
		Color temp;
		// jitter the ray randomly within a pixel
		temp.red = temp.green = temp.blue = 0;
		rayDirection = (b-eye)
		  +((i+RAND())*(a-b)/WIDTH)
		  +((j+RAND())*(c-b)/HEIGHT);
		rayDirection.normalize();
		// if we want to simulate depth of field
		if (DEPTH_OF_FIELD)
		  {
		    // compute the intersection point with the focal plane
		    focal_intersect = 
		      focal_plane->intersection(eye,rayDirection,status);
		    // if there is an intersection with the focal plane
		if (status == 1)
		  {
		    // choose a random point on the lens and trace by passing
		    // through the intersection point on the focal plane
		    lens_point[0] = eye[0]+RAND();
		    lens_point[1] = eye[1]+RAND();
		    lens_point[2] = eye[2]+RAND();
		    // create the new ray to simulate depth of field
		    rayDirection[0] = focal_intersect[0] - lens_point[0];
		    rayDirection[1] = focal_intersect[1] - lens_point[1];
		    rayDirection[2] = focal_intersect[2] - lens_point[2];
		    rayDirection.normalize();
		    temp = trace(lens_point,rayDirection,0);
		  }
		else
		  temp = trace(eye,rayDirection,0);
		  }
		else
		  // else if we miss the focal plane then the object is in 
		  // focus
		  temp = trace(eye,rayDirection,0);
		// add the color components
		pixelColor.red = pixelColor.red + temp.red; 
		pixelColor.green = pixelColor.green + temp.green; 
		pixelColor.blue = pixelColor.blue + temp.blue; 
	      }
	    // take the average of the samples we have to simulate fuzzy
	    // effects
	    pixelColor.red = pixelColor.red/(SAMPLES+1);
	    pixelColor.green = pixelColor.green/(SAMPLES+1);
	    pixelColor.blue = pixelColor.blue/(SAMPLES+1);
	    // normalize the color to avoid artifacts
	    if (pixelColor.red > 1)
	      pixelColor.red = 1;
	    if (pixelColor.green > 1)
	      pixelColor.green = 1;
	    if (pixelColor.blue > 1)
	      pixelColor.blue = 1;
	    if (pixelColor.red < 0 )
	      pixelColor.red = 0;
	    if (pixelColor.green < 0)
	      pixelColor.green = 0;
	    if (pixelColor.blue < 0)
	      pixelColor.blue = 0;
	    // update the scene buffer 
	    scene[i][j].red = scene[i][j].red + pixelColor.red;
	    scene[i][j].green = scene[i][j].green + pixelColor.green;
	    scene[i][j].blue = scene[i][j].blue + pixelColor.blue;	
	  }
      // if we want to simulate motion blur then update the location of the
      // objects in the scene 
      if (MOTION_BLUR)
	updateLocation();
      else
	break;
    }
  // dump the contents of the scene buffer into the portable pixel map file
  for (int i = 0 ; i < WIDTH ; i++)
    for ( int j = 0 ; j < HEIGHT ; j++)
      {
	output.pixel(i,j,
		     (scene[i][j].red/(TIME))*255,
		     (scene[i][j].green/(TIME))*255,
		     (scene[i][j].blue/(TIME))*255);
      }
  // done tracing
  cout<<"Ray tracing ends ...\n";
}

//-----------------------------------------------------------------------------
// Parameters    : gmVector3 p, gmVector3 direction,int level
// Return value  : Color
// Purpose       : This function is used to compute intersection test with
//                 objects in the scene and to shade the intersection point if
//                 it exists using phong shading
// Method called : trace
//-----------------------------------------------------------------------------
Color RayTracer::trace(gmVector3 p, gmVector3 direction,int level)
{
  int status = 0;
  gmVector3 temp(0,0,0);
  gmVector3 q,r,n;
  // set to the background
  Color finalColor = background;
  Color reflectColor;
  reflectColor.red = reflectColor.green = reflectColor.blue = 0;
  // check the depth of the reflection 
  if (level > MAX)
    return finalColor;
  // check intersections with all the objects in the scene
  for (unsigned int i = 0 ; i < shapes.size() ; i++)
    {
      status = 0;
      q = shapes[i]->intersection(p,direction,status);
      // if there is no intersection then check other objects
      if (status == 0)
	continue;
      else
      	{
	  // compute the normal vector at the intersetion point
	  n = shapes[i]->getNormalVector(q);
	  // compute the reflection vector at the intersection point
	  r = shapes[i]->getReflectionVector(direction,n);
	  // get the color at the intersection point using phong shading
	  finalColor = phongShading(q,n,r,i);
	  if (shapes[i]->getMat().reflection > 0)
	    {
	      // if we want to simulate blurry reflections then we need to 
	      // jitter the reflection ray randomly
	      if (BLURRED_REFLECTION)
		{
		  // jitter the refection ray to allow blurry reflections
		  temp[0] = r[0] + RAND();
		  temp[1] = r[1] + RAND();
		  temp[2] = r[2] + RAND();
		  // trace the secondary ray which is known as the 
		   // reflectin vector
		  reflectColor = trace(q,temp,level+1);
		}
	      else
		// if we want the normal reflection 
		   reflectColor = trace(q,r,level+1);
	      // compute the final color with respect to reflection
	      finalColor.red = (1-shapes[i]->getMat().reflection)
		*finalColor.red
		+ (shapes[i]->getMat().reflection*reflectColor.red);
	      finalColor.green = (1-shapes[i]->getMat().reflection)
		*finalColor.green
		+ (shapes[i]->getMat().reflection*reflectColor.green);
	      finalColor.blue = (1-shapes[i]->getMat().reflection)
		*finalColor.blue
		+ (shapes[i]->getMat().reflection*reflectColor.blue);
	      return finalColor;
	    }
	  else
	    return finalColor;
	}
    }
  // our final pixel color 
  return finalColor;
}

//-----------------------------------------------------------------------------
// Parameters    : gmVector3 p,gmVector3 l,unsigned int index
// Return value  : bool
// Purpose       : This function is used to test whether a point is visible
//                 to the light source ( either point or area light source )
// Method called : shadowFeeler
//-----------------------------------------------------------------------------
bool RayTracer::shadowFeeler(gmVector3 p,gmVector3 l,unsigned int index)
{
  gmVector3 q;
  int status = 0;
  // check intersection with all the objects except the source object
  for (unsigned int i = 0 ; i < shapes.size() ; i++)
    {
      status = 0;
      // avoid self intersection
      if (i == index)
	continue;
      else
	{
	  // check intersections with other objects in the scene
	  q = shapes[i]->intersection(p,l,status);
	  if (status == 0)
	    continue;
	  else
	    // return false if we hit something
	    return false;
	}
    }
  // return true if we don't hit somthing
  return true;
}

//-----------------------------------------------------------------------------
// Parameters    : gmVector3 intersectionPoint,gmVector3 nVec,
//		   gmVector3 rVec, unsigned int index
// Return value  : Color
// Purpose       : This function is used to calculate the intensity of a pixel
//                 using phong shading model
// Method called : phongShading
//-----------------------------------------------------------------------------
Color RayTracer::phongShading(gmVector3 intersectionPoint,
			      gmVector3 nVec,
			      gmVector3 rVec,
			      unsigned int index)
{
  // set color to the background initially
  Color result(background.red,background.green,background.blue);
  gmVector3 lVec,vVec,hVec;
  double lAngle,rAngle;
  Color tempCol;
  tempCol.red = tempCol.green = tempCol.blue = 0;
  result = tempCol;
  // go over all the light sources in our area light 
  for ( unsigned int i = 0 ; i < area_light.size() ; i++)
    {
      // calculate the light vector
      lVec = area_light[i].pos - intersectionPoint;
      lVec.normalize();
      // check if we don't hit anything while distributing the rays to either
      // the point light source or the area light source
      if (shadowFeeler(intersectionPoint,lVec,index))
	{
	  // calculation needed by the phong model
	  vVec = -1 * rayDirection;
	  vVec.normalize();
	  hVec = vVec + lVec;
	  hVec.normalize();
	  lAngle = dot(lVec,nVec);
	  if (lAngle < 0)
	    lAngle = 0;
	  result.red += area_light[i].lightMat.diffuse.red
	    * shapes[index]->getMat().diffuse.red * lAngle;
	  result.green += area_light[i].lightMat.diffuse.green
	    * shapes[index]->getMat().diffuse.green * lAngle;
	  result.blue += area_light[i].lightMat.diffuse.blue
	    * shapes[index]->getMat().diffuse.blue * lAngle;
	  rAngle = dot(hVec,nVec);
	  rAngle = pow(rAngle,shapes[index]->getMat().shininess);
	  if (rAngle < 0)
	    rAngle = 0;
	  if (rAngle > 1)
	    rAngle = 1;
	  result.red += area_light[i].lightMat.specular.red
	    * shapes[index]->getMat().specular.red * rAngle;
	  result.green += area_light[i].lightMat.specular.green
	    * shapes[index]->getMat().specular.green * rAngle;
	  result.blue += area_light[i].lightMat.specular.blue
	    * shapes[index]->getMat().specular.blue * rAngle;
	}
    }
  // take the average to simulate soft shadows
  result.red = result.red / (area_light.size());
  result.green = result.green / (area_light.size());
  result.blue = result.blue / (area_light.size());
  return result;
}

//-----------------------------------------------------------------------------
// Parameters    : None
// Return value  : None
// Purpose       : This function is used to dump everthing to a ppm file
// Method called : outputImage
//-----------------------------------------------------------------------------
void RayTracer::outputImage()
{
  // create a stream to a stream
  ofstream outfile("35.ppm");
  // dump the final scene 
  outfile<<output;
  // close the file
  outfile.close();
}





